/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.ext.feature.auth.jaas;

import java.io.Serializable;
import java.security.Principal;
import java.util.Map;

import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;

import org.bluestemsoftware.specification.eoa.ext.feature.auth.HostInfo;
import org.bluestemsoftware.specification.eoa.ext.feature.auth.cstore.Credential;
import org.bluestemsoftware.specification.eoa.ext.feature.auth.cstore.CredentialStore;
import org.bluestemsoftware.specification.eoa.ext.feature.auth.cstore.CredentialStoreFeature;
import org.bluestemsoftware.specification.eoa.ext.feature.auth.jaas.LoginCredentials;
import org.bluestemsoftware.specification.eoa.ext.feature.auth.jaas.LoginCallbackHandler.CredentialsCallback;
import org.bluestemsoftware.specification.eoa.ext.feature.http.server.HTTPAuthenticationResponse;
import org.bluestemsoftware.specification.eoa.system.System.Log;
import org.bluestemsoftware.specification.eoa.system.SystemContext;
import org.bluestemsoftware.specification.eoa.system.server.Server;

public final class LoginModuleImpl implements LoginModule {

    private static Log log = SystemContext.getContext().getSystem().getLog(LoginModuleImpl.class);

    private CredentialStore cstore;
    private boolean authenticated;
    private boolean committed;
    private Subject subject;
    private Subject privateState;
    private CallbackHandler callbackHandler;
    private HostInfo hostInfo;

    public void initialize(Subject subject, CallbackHandler callbackHandler, Map<String, ?> sharedState, Map<String, ?> options) {
        this.subject = subject;
        this.callbackHandler = callbackHandler;
        String cstoreID = null;
        if (options == null) {
            throw new IllegalStateException("Missing required options");
        }
        cstoreID = (String)options.get("cstore");
        if (cstoreID == null) {
            throw new IllegalStateException("Required option 'cstore' is undefined.");
        }
        Server server = SystemContext.getContext().getSystem().getServer();
        CredentialStoreFeature cstoreFeature = server.getFeature(CredentialStoreFeature.class);
        cstore = cstoreFeature.getCredentialStore(cstoreID);
        if (cstore == null) {
            throw new IllegalStateException("CredentialStore '" + cstoreID + "' is undefined.");
        }
        hostInfo = new HostInfo(HostInfo.MY_HOST);
    }

    public boolean login() throws LoginException {

        if (callbackHandler == null) {
            throw new LoginException("CallbackHandler instance required to authenticate.");
        }

        try {

            Callback[] callbacks = new Callback[1];
            callbacks[0] = new CredentialsCallback();
            callbackHandler.handle(callbacks);

            LoginCredentials credentials = null;
            credentials = ((CredentialsCallback)callbacks[0]).getCredentials();

            String userName = credentials.getUser();

            String realm = null;
            if (credentials instanceof HTTPAuthenticationResponse) {
                realm = ((HTTPAuthenticationResponse)credentials).getRealm();
            } else {
                realm = "default";
            }

            if ((userName == null) || (realm == null)) {
                return authenticated = false;
            }

            Credential temp = cstore.getCredential(hostInfo, realm, userName);
            if (temp == null) {
                if (log.isDebugEnabled()) {
                    log.debug("no credentials defined for user: '"
                            + userName
                            + "' within realm: '"
                            + realm
                            + "'");
                } else {
                    log.warn("auth failure: user: '"
                            + userName
                            + "' within realm: '"
                            + realm
                            + "'");
                }
                return authenticated = false;
            }

            privateState = new Subject();
            privateState.getPrincipals().add(new PrincipalImpl(userName));
            privateState.getPrivateCredentials().add(temp.getPassword());

            authenticated = temp.getPassword().check(credentials);
            
            if (!authenticated) {
                if (log.isDebugEnabled()) {
                    log.debug("invalid password for user: '"
                            + userName
                            + "' within realm: '"
                            + realm
                            + "'");
                } else {
                    log.warn("auth failure: user: '"
                            + userName
                            + "' within realm: '"
                            + realm
                            + "'");
                }
            }
            
            return authenticated;

        } catch (Exception ex) {
            throw new LoginException(ex.toString());
        }

    }

    public boolean logout() throws LoginException {
        subject.getPrincipals().removeAll(privateState.getPrincipals());
        subject.getPrivateCredentials().removeAll(privateState.getPrivateCredentials());
        return true;
    }

    public boolean abort() throws LoginException {
        privateState = null;
        return (authenticated && committed);
    }

    public boolean commit() throws LoginException {
        if (!authenticated) {
            privateState = null;
            return committed = false;
        }
        committed = true;
        subject.getPrincipals().addAll(privateState.getPrincipals());
        subject.getPrivateCredentials().addAll(privateState.getPrivateCredentials());
        return true;
    }

    static class PrincipalImpl implements Principal, Serializable {

        private static final long serialVersionUID = 1L;
        private String name = null;

        public PrincipalImpl(String userName) {
            this.name = userName;
        }

        public String getName() {
            return name;
        }

        public final String toString() {
            return name;
        }

        @Override
        public final int hashCode() {
            return hashCode(name);
        }

        @Override
        public final boolean equals(Object obj) {
            if (obj == this) {
                return true;
            }
            if (!(obj instanceof PrincipalImpl)) {
                return false;
            }
            if ((name == ((PrincipalImpl)obj).name || (name != null && name.equals(((PrincipalImpl)obj).name)))) {
                return true;
            }
            return false;
        }

        private int hashCode(Object object) {
            return (object == null ? 0 : object.hashCode());
        }

    }

}
